XPerl_GrimReaper_Config = {
	dockToTooltip = true,
	scale	= 0.6,
	keep = 15,
	bars = true,
	barsLeft = true,
}

local conf = XPerl_GrimReaper_Config

local lastTime = 0
local millisecs = 0

local gr = CreateFrame("Frame", "XPerl_GrimReaper")
gr.list = {}
gr.frameWidth = 190
--gr.timerAccuracy = 30		-- How many iterations needed to create accuracy timer, more is better

gr:SetWidth(1)
gr:SetHeight(1)

local schools = {DAMAGE_SCHOOL2, DAMAGE_SCHOOL3, DAMAGE_SCHOOL4, DAMAGE_SCHOOL5, DAMAGE_SCHOOL6, DAMAGE_SCHOOL7}
local schoolColour = {{1, 1, 0.5}, {1, 0, 0}, {0, 1, 0}, {0.3, 0.3, 1}, {0.5, 0, 0.5}, {0.5, 0.5, 0.5}}
schools[0] = SPELL_SCHOOL0_CAP
schoolColour[0] = {1, 0.5, 0.5}
local healColour = {1, 1, 1}
local baseColour = {1, 1, 1}
local specialColour = {0.5, 1, 1}

local recent = setmetatable({}, {__mode = "k"})

-- TEST ONLY:
--gr.list.Zali = {
--	{"WOUND", "", 25, 0, 1182186763, 90.121051899263, 83888.176},
--	{"WOUND", "BLOCK", 0, 0, 1182186763, 90.121051899263, 83888.659},
--	{"WOUND", "", 93, 0, 1182186765, 88.82704883818, 83890.358},
--	{"WOUND", "CRUSHING", 608, 0, 1182186767, 87.324335605955, 83892.369},
--	{"WOUND", "", 110, 0, 1182186768, 85.793794350911, 83893.612},
--	{"MISS", "", 0, 0, 1182186770, 84.277167107277, 83895.753},
--	{"WOUND", "", 109, 0, 1182186772, 84.277167107277, 83897.711},
--	{"WOUND", "", 113, 0, 1182186774, 84.277167107277, 83899.68},
--	{"WOUND", "", 100, 0, 1182186776, 81.313482677056, 83901.819},
--	{"WOUND", "", 107, 0, 1182186778, 79.82468345624, 83903.751},
--	{"WOUND", "", 103, 0, 1182186780, 77.250591345485, 83905.603},
--	{"WOUND", "", 82, 0, 1182186782, 77.626269653541, 83907.764},
--	{"HEAL", "", 836, 0, 1182186787, 89.63406149993, 83912.849},
--	{"WOUND", "", 9890, 3, 1182186790, "DEAD", 83915.608},
--}

if (not XPerlColourTable) then
	XPerlColourTable = setmetatable({},{
		__mode = 'k',
		__index = function(self, class)
			local c = RAID_CLASS_COLORS[strupper(class)] or {r = 0, g = 0, b = 0}
			self[class] = format("|c00%02X%02X%02X", 255 * c.r, 255 * c.g, 255 * c.b)
			return self[class]
		end
	})
end

-- gr:OnEvent(event, a, b, c, d)
function gr:OnEvent(event, a, b, c, d, e)
	self[event](self, a, b, c, d, e)
end

-- AddCombat
function gr:AddCombat(unit, dmgType, modifier, amount, school)
	if (dmgType == "WOUND" and amount == 0) then
		return
	end

	local name = UnitName(unit)

	local list = self.list[name]
	if (not list) then
		list = {}
		self.list[name] = list
	end

	-- Note that UnitHealth is not updated immediately at this point, so we setup the 'recent' table to look for UNIT_HEALTH
	-- event that comes for this unit shortly when we can reliably measure the unit's health

	local health
	if (UnitIsDeadOrGhost(unit)) then
		health = "DEAD"
	else
		health = floor(100 * UnitHealth(unit) / UnitHealthMax(unit))
	end

	local new = {dmgType, modifier, amount, school, time(), health, floor(millisecs * 1000)}
	tinsert(list, new)
	recent[unit] = new

	while (#list > conf.keep) do
		tremove(list, 1)
	end

	if (self.lastUnit and UnitIsUnit(unit, self.lastUnit)) then
		self:Tip()		-- Refresh if shown
	end
end

-- UNIT_COMBAT
-- We assign UNIT_COMBAT for whichever group type we're in, UNIT_COMBAT is called a LOT so I wanted this as fast as possible

function gr:UNIT_COMBAT_raid(unit, dmgType, modifier, amount, school)
	if (strfind(unit, "^raid%d+")) then
		self:AddCombat(unit, dmgType, modifier, amount, school)
	end
end

function gr:UNIT_COMBAT_party(unit, dmgType, modifier, amount, school)
	if (strfind(unit, "^party%d") or unit == "player") then
		self:AddCombat(unit, dmgType, modifier, amount, school)
	end
end

-- function gr:UNIT_COMBAT(unit, dmgType, modifier, amount, school)
-- 	local ok
-- 	if (GetNumRaidMembers() > 0) then
-- 		if (strfind(unit, "^raid%d+")) then
-- 			ok = true
-- 		end
-- 	else
-- 		if (strfind(unit, "^party%d") or unit == "player") then
-- 			ok = true
-- 		end
-- 	end
--
-- 	if (ok) then
-- 		self:AddCombat(unit, dmgType, modifier, amount, school)
-- 	end
-- end

-- gr:UNIT_HEALTH()
function gr:UNIT_HEALTH(unit)
	local r = recent[unit]
	if (r) then
		if (UnitIsDeadOrGhost(unit)) then
			r[6] = "DEAD"
		else
			r[6] = floor(100 * UnitHealth(unit) / UnitHealthMax(unit))
		end
		recent[unit] = nil

		if (unit == self.lastUnit) then
			self:Tip()		-- Refresh if shown
		end
	end
end

-- RAID_ROSTER_UPDATE
function gr:RAID_ROSTER_UPDATE()
	self:Validate()
end

-- PARTY_MEMBERS_CHANGED
function gr:PARTY_MEMBERS_CHANGED()
	if (GetNumRaidMembers() == 0) then
		-- Don't do this if in a raid, no need
		self:Validate()
	end
end

-- Validate
function gr:Validate()

	-- Clear out members of list that have left raid
	local current = {}
	if (GetNumRaidMembers() > 0) then
		for i = 1,GetNumRaidMembers() do
			local name = GetRaidRosterInfo(i)
			if (name) then
				current[name] = true
			end
		end

		self.UNIT_COMBAT = self.UNIT_COMBAT_raid
	else
		current[UnitName("player")] = true
		for i = 1,GetNumPartyMembers() do
			local name = UnitName("party"..i)
			if (name) then
				current[name] = true
			end
		end

		self.UNIT_COMBAT = self.UNIT_COMBAT_party
	end

	for k,v in pairs(self.list) do
		if (not current[k]) then
			self.list[k] = nil
		end
	end
end

-- gr:Timer(elapsed)
-- This gives us a time accuracy of 100th seconds. The actual time is not important, only the time 'between' events, so this will do
function gr:Timer(elapsed)
	local t = time()
	if (t > lastTime) then
		lastTime = t
		millisecs = 0
	else
		millisecs = millisecs + elapsed
		if (millisecs >= 1) then
			millisecs = 0.999
		end
	end
end

-- PLAYER_ENTERING_WORLD
function gr:PLAYER_ENTERING_WORLD()

	self.UNIT_COMBAT = self.UNIT_COMBAT_party

	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_COMBAT")
	self:RegisterEvent("UNIT_HEALTH")

	hooksecurefunc(GameTooltip, "SetUnit", self.TipHook_SetUnit)

	self:SetScript("OnUpdate", self.Timer)
end

-- VARIABLES_LOADED
function gr:VARIABLES_LOADED()
	conf = XPerl_GrimReaper_Config

	if (conf.bars == nil) then
		conf.bars = true
	end
	if (conf.barsLeft) then
		conf.barsLeft = true
	end

	conf.scale	= conf.scale	or 0.6
	conf.keep	= conf.keep	or 15

	self:RestorePosition()
end

-- TipHook_OnShow
function gr:TipHook_SetUnit(unit)
	XPerl_GrimReaper:Tip(unit, self)
end

-- gr:FixLastUnit()
function gr:FixUnit(unit)
	if (GetNumRaidMembers() > 0) then
		if (not strfind(unit, "^raid")) then
			for i = 1,GetNumRaidMembers() do
				if (UnitIsUnit("raid"..i, unit)) then
					return "raid"..i
				end
			end
		end

	else
		if (UnitIsUnit("player", unit)) then
			return "player"
		end

		if (GetNumPartyMembers() > 0) then
			if (not strfind(unit, "^party")) then
				for i = 1,GetNumRaidMembers() do
					if (UnitIsUnit("party"..i, unit)) then
						return "party"..i
					end
				end
			end
		end
	end

	return unit
end

-- Tip
function gr:Tip(unit, tooltip)

	if (not unit) then
		unit = self.lastUnit
		tooltip = self.lastTip
	end

	local att = self.attachment
	local unitName = UnitName(unit)
	local list = self.list[unitName]

	if (list) then
		local lastTime = list[#list][5]

		unit = self:FixUnit(unit)
		self.lastUnit = unit
		self.lastTip = tooltip

		if (not att) then
			self.attachment = CreateFrame("Frame", "XPerl_GrimReaper_Attachment", tooltip)
			att = self.attachment
			att:SetScale(conf.scale)

			if (conf.dockPoint == 1) then
				att:SetPoint("BOTTOMLEFT", tooltip, "BOTTOMRIGHT", 0, 0)
			elseif (conf.dockPoint == 2) then
				att:SetPoint("BOTTOMRIGHT", tooltip, "BOTTOMLEFT", 0, 0)
			end

			att:SetWidth(self.frameWidth)
			att:SetHeight(100)
			att:SetClampedToScreen(true)

			local bgDef = {bgFile = "Interface\\Addons\\XPerl\\Images\\XPerl_FrameBack",
	       				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	       				tile = true, tileSize = 32, edgeSize = 24,
	       				insets = { left = 6, right = 6, top = 6, bottom = 6 }
				}
			att:SetBackdrop(bgDef)
			att:SetBackdropColor(0, 0, 0, 0.8)
			att:SetBackdropBorderColor(1, 1, 1, 1)

			att.title = att:CreateFontString(nil, "BORDER", "GameFontNormal")
			att.title:SetText(XPerl_ShortProductName.." "..XPERL_REAPER_TITLE)
			att.title:SetPoint("BOTTOM", att, "TOP", 0, 0)
			att.title:SetHeight(15)
			att.title:SetWidth(self.frameWidth)
			att.title:SetTextColor(1, 1, 1)

			att.frames = {}

			tooltip:HookScript("OnHide",
				function(self)
					if (conf.dockToTooltip) then
						XPerl_GrimReaper_Attachment:Hide()
						XPerl_GrimReaper.lastUnit = nil
						XPerl_GrimReaper.lastTip = nil
					end
				end)

			att:SetScript("OnDragStart",
				function(self)
					if (not conf.dockToTooltip) then
						XPerl_GrimReaper:StartMoving()
					end
				end)

			att:SetScript("OnDragStop",
				function(self)
					XPerl_GrimReaper:StopMovingOrSizing()
					XPerl_GrimReaper:SavePosition()
					XPerl_GrimReaper:Tip()
				end)

			att:SetScript("OnMouseUp",
				function(self, button)
					if (button == "RightButton") then
						XPerl_GrimReaper:ShowMenu()
					end
				end)

			att:SetScript("OnMouseWheel",
				function(self)
					if (IsAltKeyDown()) then
						conf.scale = conf.scale + (0.05 * arg1)
						if (conf.scale < 0.2) then
							conf.scale = 0.2
						elseif (conf.scale > 2) then
							conf.scale = 2
						end
						att:SetScale(conf.scale)
					end
				end)
		end

		att:ClearAllPoints()
		if (conf.dockToTooltip) then
			att:SetParent(tooltip)
			att:SetPoint("BOTTOMRIGHT", tooltip, "BOTTOMLEFT", 0, 0)
			self:SetMovable(false)
			att:EnableMouseWheel(false)
			att:EnableMouse(false)
		else
			att:SetParent(UIParent)
			att:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
			self:SetMovable(true)
			att:EnableMouse(true)
			att:EnableMouseWheel(true)
			att:RegisterForDrag("LeftButton")
		end

		local _, unitClass = UnitClass(unit)
		att.title:SetText(XPerl_ShortProductName.." "..XPERL_REAPER_TITLE.." - "..XPerlColourTable[unitClass]..unitName)

		local i = 1
		for k = #list,1,-1 do
			local v = list[k]
			local frame = att.frames[i]
			if (not frame) then
				frame = CreateFrame("Frame", "XPerl_GrimReaper_Attachment"..k, att)
				att.frames[i] = frame

				if (i == 1) then
					frame:SetPoint("BOTTOMRIGHT", att, "BOTTOMRIGHT", -6, 10)
				else
					frame:SetPoint("BOTTOMRIGHT", att.frames[i - 1], "TOPRIGHT", 0, 0)
				end

				frame:SetWidth(self.frameWidth - 10)
				frame:SetHeight(15)
				--frame:SetAlpha(0.8)

				frame.text = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
				frame.text:SetJustifyH("RIGHT")
				frame.text:SetAllPoints(frame)

				frame.time = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
				frame.time:SetJustifyH("LEFT")
				frame.time:SetAllPoints(frame)
				frame.time:SetTextColor(1, 1, 0.5)
			end

			if (conf.bars) then
				if (not frame.texture) then
					frame.texture = frame:CreateTexture(nil, "BORDER")
					frame.texture:SetTexture(XPerl_GetBarTexture())
					frame.texture:SetWidth(1)
					frame.texture:SetHeight(15)

					frame.healthText = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
					frame.healthText:SetWidth(90)
					frame.healthText:SetTextColor(1, 1, 1)
				end

				frame.texture:ClearAllPoints()
				frame.healthText:ClearAllPoints()
				if (conf.barsLeft) then
					frame.texture:SetPoint("RIGHT", frame.text, "LEFT", -4, 0)
					frame.healthText:SetJustifyH("RIGHT")
					frame.healthText:SetPoint("RIGHT", frame.text, "LEFT", -7, 1)
				else
					frame.texture:SetPoint("LEFT", frame.text, "RIGHT", 4, 0)
					frame.healthText:SetJustifyH("LEFT")
					frame.healthText:SetPoint("LEFT", frame.text, "RIGHT", 7, 1)
				end
				frame.texture:Show()
				frame.healthText:Show()
			else
				if (frame.texture) then
					frame.texture:Hide()
					frame.healthText:Hide()
				end
			end

			local str
			local color = schoolColour[v[4]] or basicColour

			if (XPERL_REPEAR_SPECIALS[v[1]]) then
				color = specialColour
				str = XPERL_REPEAR_SPECIALS[v[1]]

			elseif (v[2] and XPERL_REPEAR_SPECIALS[v[2]]) then
				color = specialColour
				str = XPERL_REPEAR_SPECIALS[v[2]]

			elseif (v[1] == "HEAL") then
				color = healColour
				str = "+"..tostring(v[3])
			else
				str = "-"..tostring(v[3])
			end

			if (v[2] == "CRUSHING") then
				str = XPERL_REAPER_CRUSHING.." "..str
			end

			frame.text:SetText(str)
			frame.text:SetTextColor(color[1], color[2], color[3])

			if (v[2] == "CRITICAL" or v[2] == "CRUSHING") then
				frame.text:SetFontObject("GameFontNormalLarge")
			else
				frame.text:SetFontObject("GameFontNormal")
			end

			frame.time:SetText(date(XPERL_REAPER_TIMEFORMAT, v[5]).."|c00C0FFC0"..self:MakeFractionalTime(v[7]))

			if (v[6] == "DEAD") then
				frame.healthText:SetText(DEAD)
				frame.texture:SetWidth(1)
				frame.texture:SetVertexColor(1, 0, 0)
			else
				frame.healthText:SetText(format("%d"..PERCENT_SYMBOL, v[6]))
				frame.texture:SetWidth(v[6])

				local r, g, b
				local percentage = v[6] / 100
				if (percentage < 0.5) then
					r = 1
					g = 2*percentage
					b = 0
				else
					g = 1
					r = 2*(1 - percentage)
					b = 0
				end

				if (r >= 0 and g >= 0 and b >= 0 and r <= 1 and g <= 1 and b <= 1) then
					frame.texture:SetVertexColor(r, g, b)
				end
			end

			frame:Show()
			i = i + 1
		end
		for i = #list + 1,15 do
			local frame = att.frames[i]
			if (frame) then
				frame:Hide()
			end
		end

		att:SetHeight(15 * #list + 17)
		self:SetHeight(15 * #list + 17)
		att:Show()
	else
		if (att and conf.dockToTooltip) then
			att:Hide()
		end
	end
end

-- SavePosition
function gr:SavePosition()
	if (self.attachment) then
		conf.SavedPosition = {top = self:GetTop(), left = self:GetLeft()}
	end
end

-- RestorePosition
function gr:RestorePosition()
	local pos = conf.SavedPosition
	if (pos) then
		if (pos.left or pos.right) then
			self:ClearAllPoints()
			self:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", pos.left, pos.top)
		end
	end
end

-- Output
function gr:Output(str)
	if (GetNumRaidMembers() > 0) then
		SendChatMessage(str, "RAID")
	elseif (GetNumPartyMembers() > 0) then
		SendChatMessage(str, "PARTY")
	else
		self:Msg(str)
	end
end

-- self:Msg(str)
function gr:Msg(str)
	DEFAULT_CHAT_FRAME:AddMessage("|c00FFFF80"..str)
end

-- MakeFractionalTime
function XPerl_GrimReaper:MakeFractionalTime(getTime)
	if (getTime) then
		return format(".%03d", getTime)
	end
	return ""
end

-- Report
function gr:Report(unitName)

	local list = self.list[unitName]

	if (list) then
		local lastTime = list[#list][5]

		self:Output(format(XPERL_REPEAR_REPORTTITLE, unitName))

		for k,v in ipairs(list) do
			local str = date(XPERL_REAPER_TIMEFORMAT, v[5])..self:MakeFractionalTime(v[7]).."> "

			if (XPERL_REPEAR_SPECIALS[v[1]]) then
				str = str..XPERL_REPEAR_SPECIALS[v[1]]

			elseif (XPERL_REPEAR_SPECIALS[v[2]]) then
				str = str..XPERL_REPEAR_SPECIALS[v[2]]

			elseif (v[1] == "HEAL") then
				str = str..XPERL_REPEAR_REPORTHEAL.." +"..tostring(v[3])

			else
				str = str.."-"..tostring(v[3])
			end

			if (v[2] ~= "") then
				if (v[2] == "CRUSHING") then
					str = str.." ("..XPERL_REPEAR_REPORTCRUSH..")"
				elseif (v[2] == "CRITICAL") then
					str = str.." ("..XPERL_REPEAR_REPORTCRIT..")"
				end
			end

			if (v[4] ~= 0 and schools[v[4]]) then
				str = str.." ["..schools[v[4]].."]"
			end

			self:Output(str)
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage(format(XPerl_ShortProductName.." "..XPERL_REAPER_TITLE..": |c00FFFF80"..XPERL_REPEAR_REPORT_NF, unitName))
	end
end

-- ShowMenu
function gr:ShowMenu()

	if (not self.menu) then
		self.menu = CreateFrame("Frame", "XPerl_GrimReaper_Menu")
		self.menu.displayMode = "MENU"
		UIDropDownMenu_Initialize(self.menu, function(level) XPerl_GrimReaper:Menu_Initialize(level) end)
	end

	ToggleDropDownMenu(1, nil, XPerl_GrimReaper_Menu, "cursor")
end

-- ShowMenu
function gr:Menu_Initialize(level)

	local info

	info = {}
	info.isTitle = 1
	info.text = XPERL_REAPER_TITLE
	UIDropDownMenu_AddButton(info)

	info = {}
	info.func = function()
			conf.bars = not conf.bars
			XPerl_GrimReaper:Tip()
		end
	info.checked = conf.bars
	info.text = XPERL_REAPER_MENU_BARS
	UIDropDownMenu_AddButton(info)

	info = {}
	info.func = function()
			conf.barsLeft = not conf.barsLeft
			XPerl_GrimReaper:Tip()
		end
	info.checked = conf.barsLeft
	info.text = XPERL_REAPER_MENU_BARSLEFT
	UIDropDownMenu_AddButton(info)

	info = {}
	info.func = function()
			conf.dockToTooltip = not conf.dockToTooltip
			XPerl_GrimReaper:Tip()
		end
	info.text = XPERL_REAPER_MENU_DOCK
	UIDropDownMenu_AddButton(info)

	info = {}
	info.func = function()
			XPerl_GrimReaper.attachment:Hide()
			XPerl_GrimReaper.lastUnit = nil
			XPerl_GrimReaper.lastTip = nil
		end
	info.text = XPERL_REAPER_MENU_HIDE
	UIDropDownMenu_AddButton(info)
end

-- OnOff
local function OnOff(onoff)
	if (onoff) then
		return "("..XPERL_REAPER_ON_CLR..XPERL_REAPER_ON.."|r)"
	end
	return "("..XPERL_REAPER_OFF_CLR..XPERL_REAPER_OFF.."|r)"
end

-- SlashCmd
function gr:SlashCmd(msg)

	if (not msg or msg == "") then
		for i = 1,100 do
			local text = getglobal("XPERL_REAPER_HELP"..i)
			if (not text) then
				break
			end

			if (strfind(text, XPERL_REAPER_CMD_KEEP)) then
				text = format(text, conf.keep)
			elseif (strfind(text, XPERL_REAPER_CMD_SCALE)) then
				text = format(text, conf.scale)
			end

			DEFAULT_CHAT_FRAME:AddMessage(text)
		end
		return
	end

	local keep = tonumber(strmatch(strlower(msg), "^"..XPERL_REAPER_CMD_KEEP.." (%d+)"))
	if (keep) then
		if (keep >= 1 and keep <= 50) then
			conf.keep = keep
			self:Msg(format(XPERL_REAPER_KEEPING, keep))
		end
		return
	end

	local scale = tonumber(strmatch(strlower(msg), "^"..XPERL_REAPER_CMD_SCALE.." (%f+)"))
	if (scale) then
		if (scale >= 0.2 and scale <= 2) then
			conf.scale = scale
			XPerl_GrimReaper:SetScale(conf.scale)
		end
		return
	end

	if (strlower(msg) == XPERL_REAPER_CMD_DOCK) then
		conf.dockToTooltip = not conf.dockToTooltip
		self:Msg(format(XPERL_REAPER_DOCKING, OnOff(conf.dockToTooltip)))
		return
	end

	local dock = tonumber(strmatch(strlower(msg), "^"..XPERL_REAPER_CMD_DOCKP.." (%d+)"))
	if (dock) then
		conf.dockPoint = dock
		self:Msg(format(XPERL_REAPER_DOCKING, OnOff(conf.dockToTooltip)))
		return
	end

	if (strlower(msg) == XPERL_REAPER_CMD_BARS) then
		conf.bars = not conf.bars
		self:Msg(format(XPERL_REAPER_BARSET, OnOff(conf.bars)))
		return
	end

	self:Report(msg)
end

SlashCmdList["XPERL_GRIMREAPER"] = function(msg) XPerl_GrimReaper:SlashCmd(msg) end
SLASH_XPERL_GRIMREAPER1 = "/grim"

gr:SetScript("OnEvent", gr.OnEvent)
gr:RegisterEvent("PLAYER_ENTERING_WORLD")
gr:RegisterEvent("VARIABLES_LOADED")
